# -*- coding: utf-8 -*-
import pika
import json
import logging

from django.conf import settings

from pika.adapters.tornado_connection import TornadoConnection

logger = logging.getLogger("pika")
logger.setLevel(logging.INFO)


class PikaClient(object):

    def __init__(self, io_loop):
        logger.info('Initializing')
        self.io_loop = io_loop

        self.connected = False
        self.connecting = False
        self.connection = None
        self.channel = None

        self.event_listeners = set([])

    def connect(self):
        if self.connecting:
            logger.info('Already connecting to RabbitMQ')
            return

        logger.info('Connecting to RabbitMQ')
        self.connecting = True

        cred = pika.PlainCredentials(
            settings.BAGOU.get('AMQP_BROKER_USER'),
            settings.BAGOU.get('AMQP_BROKER_PASS'))
        param = pika.ConnectionParameters(
            host=settings.BAGOU.get('AMQP_BROKER_ADDR'),
            port=settings.BAGOU.get('AMQP_BROKER_PORT'),
            virtual_host=settings.BAGOU.get('AMQP_BROKER_PATH'),
            credentials=cred)

        self.connection = TornadoConnection(param, on_open_callback=self.on_connected)
        self.connection.add_on_close_callback(self.on_closed)

    def on_connected(self, connection):
        logger.info('Connected to RabbitMQ')
        self.connected = True
        self.connection = connection
        self.connection.channel(self.on_channel_open)

    def on_channel_open(self, channel):
        logger.info('Channel open, Declaring exchange')
        self.channel = channel
        self.channel.queue_declare(
            queue=settings.BAGOU.get('QUEUE_NAME'),
            durable=True,
            exclusive=False,
            auto_delete=True,
            callback=self.on_queue_declared)

    def on_queue_declared(self, frame):
        self.channel.basic_consume(self.on_message, queue=settings.BAGOU.get('QUEUE_NAME'))

    def on_closed(self, connection):
        logger.info('RabbitMQ connection closed')
        self.io_loop.stop()

    def on_message(self, channel, method, header, body):
        logger.debug('Message received: %s' % body)
        self.notify_listeners(body)

    def notify_listeners(self, event_obj):
        event_json = json.loads(event_obj)
        channels = event_json.get('channel')

        for channel in channels:
            for listener in self.event_listeners:
                if channel:
                    if channel in listener.channels:
                        listener.write_message(event_obj)
                        logger.info('Notified %s (channels: %s)' % (repr(listener), listener.channels))
                else:
                    listener.write_message(event_obj)
                    logger.info('Notified %s' % repr(listener))

    def add_event_listener(self, listener):
        self.event_listeners.add(listener)
        logger.info('Listener %s added' % repr(listener))

    def remove_event_listener(self, listener):
        try:
            self.event_listeners.remove(listener)
            logger.info('Listener %s removed' % repr(listener))
        except KeyError:
            pass



# -*- coding: utf-8 -*-
import re
import logging


class EventError(Exception):
    pass


class Event(object):
    """
    Signal-like object for Websocket events that supports
    filtering on channels. Registering event handlers is
    performed by using the Event instance as a decorator::

    @on_message
    def message(handler, message):
    ...

    Event handlers can also be registered for particular
    channels using the channel keyword argument with a
    regular expression pattern::

    @on_message(channel="^room-")
    def message(handler, message):
    ...

    The ``on_connect`` event cannot be registered with a
    channel pattern since channel subscription occurs
    after a connection is established.
    """
    def __init__(self, supports_channels=True):
        self.supports_channels = supports_channels
        self.handlers = []

    def __call__(self, handler=None, channel=None):
        """
        Decorates the given handler. The event may be called
        with only a channel argument, in which case return a
        decorator with the channel argument bound.
        """
        if handler is None:
            def handler_with_channel(handler):
                return self.__call__(handler, channel)
            return handler_with_channel
        if channel:
            if not self.supports_channels:
                raise EventError(
                    "The %s event does not support channels so "
                    "the handler `%s` could not be registered" % (
                        self.name, handler.__name__))
            channel = re.compile(channel)
        self.handlers.append((handler, channel))

    def send(self, client, message=None, *args):
        """
        When an event is sent, run all relevant handlers. Relevant
        handlers are those without a channel pattern when the given
        socket is not subscribed to any particular channel, or the
        handlers with a channel pattern that matches any of the
        channels that the given socket is subscribed to.

        In the case of subscribe/unsubscribe, match the channel arg
        being sent to the channel pattern.
        """
        if message:
            callback = message.get('callbackId')
        else:
            callback = None

        # Call at least callback if no handler set
        if message and not self.handlers and self.name is not 'on_authenticate':
            if callback:
                client.jsonify(callbackId=callback, event='callback')

        for handler, pattern in self.handlers:
            # Simple handler for open and close
            if self.name in ['on_open', 'on_close']:
                handler(client, callback)
                continue

            if self.name is 'on_authenticate':
                session_id = message.get('sessionid')
                handler(client, session_id, callback)
                continue

            no_channel = not pattern and not client.channels
            if self.name.endswith("subscribe") and pattern:
                logging.error(args)
                matches = [pattern.match(args[0].name)]
            else:
                matches = [pattern.match(c) for c in client.channels if pattern]

            if no_channel or filter(None, matches):
                channel = client.channels
                callback = None
                if message:
                    callback = message.get('callbackId')
                handler(client, channel, message, callback)

on_message = Event()
on_callback = Event()
on_subscribe = Event()
on_unsubscribe = Event()
on_open = Event(supports_channels=False)
on_close = Event(supports_channels=False)
on_store = Event(supports_channels=False)
on_authenticate = Event(supports_channels=False)

# Give each event a name attribute.
for k in list(globals()):
    v = globals()[k]
    if isinstance(v, Event):
        setattr(v, "name", k)



# -*- coding: utf-8 -*-
import json
import pika

from django.conf import settings

from ..channel import Channel


def __send(message):
    cred = pika.PlainCredentials(
        settings.BAGOU.get('AMQP_BROKER_USER'),
        settings.BAGOU.get('AMQP_BROKER_PASS'))

    param = pika.ConnectionParameters(
        host=settings.BAGOU.get('AMQP_BROKER_ADDR'),
        port=settings.BAGOU.get('AMQP_BROKER_PORT'),
        virtual_host=settings.BAGOU.get('AMQP_BROKER_PATH'),
        credentials=cred)

    body = json.dumps(message)

    conn = pika.BlockingConnection(param)
    channel = conn.channel()
    channel.basic_publish(
        exchange='', routing_key=settings.BAGOU.get('QUEUE_NAME'), body=body)


def broadcast(channels, event, data={}, callback=None):
    """
    Broadcast message to channel(s).

    channels    (str)(list) : Channel(s) to send message.
    event       (str)       : Message event type.
    data        (dict)      : Data to be attached in message.
    callback    (str)       : Callback to be send in message.
    """
    if not isinstance(channels, list) and not isinstance(channels, dict):
        channels = [channels]

    serializable_channels = []
    for channel in channels:
        serializable_channels.append(channel)

    data = {
        'event': event,
        'callbackId': callback,
        'channel': serializable_channels,
        'data': data}
    __send(data)



# -*- coding: utf-8 -*-
from django import template
from django.conf import settings

register = template.Library()


@register.inclusion_tag("bagou_scripts.html", takes_context=True)
def bagou_static(context):
    context['MEDIA_URL'] = settings.MEDIA_URL
    context['STATIC_URL'] = settings.STATIC_URL
    context['WEBSOCKET_URL'] = settings.BAGOU.get('WEBSOCKET_URL')
    return context



# -*- coding: utf-8 -*-
import logging
import tornado.web
import tornado.ioloop
import tornado.template

from django.conf import settings
from django.utils.importlib import import_module

from .client import PikaClient

logging.basicConfig()
logger = logging.getLogger("tornado.general")
logger.setLevel(logging.INFO)


class WebSocketServer(object):
    def __init__(self):
        self.io_loop = tornado.ioloop.IOLoop.instance()
        self.pika_client = PikaClient(self.io_loop)

        self.handler_module = import_module(
            '.'.join(settings.BAGOU.get('DEFAULT_HANDLER_CLASS').split('.')[:-1]))
        self.handler_class = getattr(
            self.handler_module,
            settings.BAGOU.get('DEFAULT_HANDLER_CLASS').split('.')[-1])

        self.application = tornado.web.Application()
        self.application.add_handlers(r'.*', [
            (r"%s" % settings.BAGOU.get('WEBSOCKET_PATH'), self.handler_class)])
        self.application.pika_client = self.pika_client

        self.hostname = "%s:%s" % (
            settings.BAGOU.get('WEBSOCKET_ADDR'), settings.BAGOU.get('WEBSOCKET_PORT'))

    def run(self):
        logger.info('Listening on %s:%s...' % (
            settings.BAGOU.get('WEBSOCKET_ADDR'),
            settings.BAGOU.get('WEBSOCKET_PORT')))

        self.application.channels = {}
        self.application.pika_client.connect()
        self.application.listen(int(settings.BAGOU.get('WEBSOCKET_PORT')))
        self.io_loop.start()

    def stop(self):
        for websocket in self.pika_client.event_listeners:
            websocket.close()
        self.io_loop.stop()



# -*- coding: utf-8 -*-
import json
import logging
import tornado.websocket

from django.core import serializers
from django.contrib.auth.models import User
from django.contrib.auth.models import AnonymousUser
from django.contrib.sessions.models import Session

from .events import on_open
from .events import on_close
from .events import on_store
from .events import on_message
from .events import on_subscribe
from .events import on_unsubscribe
from .events import on_authenticate

from .channel import Channel

logging.basicConfig()
logger = logging.getLogger("tornado.handler")
logger.setLevel(logging.INFO)


class WebSocketHandler(tornado.websocket.WebSocketHandler):
    def open(self, *args, **kwargs):
        self.channels = {}
        self.user = AnonymousUser()

        self.store = {}

        self.application.pika_client.add_event_listener(self)

        on_open.send(self)
        logger.info("WebSocket opened")

    def on_close(self):
        logger.info("WebSocket closed")
        self.application.pika_client.remove_event_listener(self)
        on_close.send(self)

    def _on_authenticate(self, message):
        session_id = None
        data = {}

        cookies = self.request.headers.get('Cookie')
        if cookies:
            logger.info('Cookies: ' + cookies);
            for cookie in cookies.split(';'):
                if cookie.strip().startswith('sessionid='):
                    session_id = cookie.split('=')[1].strip()
                    break

        if session_id:
            data = {}
            try:
                session = Session.objects.get(session_key=session_id)
                uid = session.get_decoded().get('_auth_user_id')
                user = User.objects.get(pk=uid)
                if user:
                    self.user = user
                    self.authenticated = True
                    logger.info('User authenticated')

                    user_json = serializers.serialize('json', [self.user])[0]
                    data.update({'user': user_json})
                else:
                    logger.info('User failed to authenticate.')
            except (Session.DoesNotExist, User.DoesNotExist):
                logger.warning('User send bad sessionid.')

            self.jsonify(
                event='callback',
                data=data,
                callbackId=message.get('callbackId'))
            on_authenticate.send(self, session_id)

    def _on_subscribe(self, message):
        channel_name = message.get('data', {}).get('channel')
        callback_id = message.get('callbackId')
        if channel_name:
            logger.info("Subscribing to '%s' channel." % channel_name)
            if channel_name not in self.application.channels:
                channel = Channel(name=channel_name, owner=self)
                self.application.channels[channel_name] = channel
            else:
                channel = self.application.channels[channel_name]
                self.application.channels[channel_name].add_client(self)

            self.channels[channel_name] = channel

            logger.info('Calling on_subscribe event handlers')
            on_subscribe.send(self, message, channel, callback_id)

    def _on_unsubscribe(self, message):
        channel_name = message.get('data', {}).get('channel')
        callback_id = message.get('callbackId')
        if channel_name:
            logger.info("Unsubscribing to '%s' channel." % channel_name)
            if channel_name in self.application.channels:
                channel = self.application.channels[channel_name]
                if self in channel.clients:
                    channel.remove_client(self)
                    del self.channels[channel_name]

                    if not channel.client and not self.persistent:
                        del self.application.channels[channel_name]

                logger.info('Calling on_unsubscribe event handlers')
                on_unsubscribe.send(self, message, channel, callback_id)

    def _on_store(self, message):
        callback_id = message.get('callbackId')
        self.store.update(message.get('data'))
        on_store.send(self, message, callback_id)

    def on_message(self, message):
        logger.info("Message received: %s " % message)

        message_json = json.loads(message)
        if not isinstance(message_json, dict):
            self.jsonify(event="error", data={'message': 'bad message format'})

        message_event = message_json.get('event')
        if message_event == 'subscribe':
            self._on_subscribe(message_json)
        elif message_event == 'unsubscribe':
            self._on_unsubscribe(message_json)
        elif message_event == 'store':
            self._on_store(message_json)
        elif message_event == 'authenticate':
            self._on_authenticate(message_json)
        else:
            logger.info('Calling on_message event handlers')
            on_message.send(self, message_json)

    def jsonify(self, **kwargs):
        self.write_message(kwargs)



# -*- coding: utf-8 -*-
import sys
import logging

try:
    from django.utils.importlib import import_module
    from django.conf import settings as django_settings
    import_events_module = True
except ImportError:
    import settings as django_settings
    import_events_module = False

if sys.version_info[0] == 3:
    from urllib.parse import urlparse
else:
    from urlparse import urlparse


__version__ = "0.1.1"

logging.basicConfig()
logger = logging.getLogger('bagou.general')
logger.setLevel(logging.INFO)

if not hasattr(django_settings, 'BAGOU'):
    settings = {}
    setattr(django_settings, 'BAGOU', settings)
else:
    settings = django_settings.BAGOU

# Default events handler
settings.setdefault('DEFAULT_HANDLER_CLASS', 'bagou.handler.WebSocketHandler')

# Websocket
settings.setdefault('WEBSOCKET_URL', 'ws://localhost:9000/websocket')
__websocket_url = urlparse(settings.get('WEBSOCKET_URL'))
settings['WEBSOCKET_ADDR'] = __websocket_url.hostname
settings['WEBSOCKET_PORT'] = int(__websocket_url.port)
settings['WEBSOCKET_PATH'] = __websocket_url.path

# AMQP
settings.setdefault('AMQP_BROKER_URL', 'amqp://guest:guest@localhost:5672/')
settings.setdefault('QUEUE_NAME', 'websocket')
__amqp_url = urlparse(settings.get('AMQP_BROKER_URL'))
settings['AMQP_BROKER_USER'] = __amqp_url.username
settings['AMQP_BROKER_PASS'] = __amqp_url.password
settings['AMQP_BROKER_ADDR'] = __amqp_url.hostname
settings['AMQP_BROKER_PORT'] = int(__amqp_url.port)
settings['AMQP_BROKER_PATH'] = __amqp_url.path

# Authentication
settings.setdefault('AUTH', True)

# Try and import an ``events`` module in each installed app,
# to ensure all event handlers are connected.
if import_events_module:
    for app in django_settings.INSTALLED_APPS:
        try:
            import_module("%s.events" % app)
            logger.info('events module found in %s' % app)
        except ImportError as err:
            pass



# -*- coding: utf-8 -*-
import uuid

from .exceptions import BagouChannelException


class Channel(object):
    def __init__(self, name=None, owner=None, allow_anonymous=True, persistent=False):
        self.name = name
        if name is None:
            self.name = str(uuid.uuid4())

        self.clients = [owner]
        self.allowed_clients = []
        self.max_client = None
        self.allow_anonymous = allow_anonymous
        self.owner = owner
        self.persistent = persistent

    def add_client(self, client):
        if client.user.is_anonymous() and not self.allow_anonymous:
            raise BagouChannelException("Anonymous users not allowed.")
        if self.allowed_clients:
            if client not in self.allowed_clients:
                raise BagouChannelException("Client not in allowed_clients.")
        if self.max_client is not None and len(self.clients) + 1 > self.max_client:
            raise BagouChannelException("Channel is full.")

        if client not in self.clients:
            self.clients.append(client)

    def remove_client(self, client):
        if client in self.clients:
            self.clients.remove(client)



# -*- coding: utf-8 -*-
import re
import sys
import socket
from datetime import datetime
from optparse import make_option
from django.core.management.base import BaseCommand
from django.core.management.base import CommandError

from django.conf import settings

from bagou import __version__
from bagou.server import WebSocketServer


naiveip_re = re.compile(r"""^(?:
(?P<addr>
    (?P<ipv4>\d{1,3}(?:\.\d{1,3}){3}) |         # IPv4 address
    (?P<ipv6>\[[a-fA-F0-9:]+\]) |               # IPv6 address
    (?P<fqdn>[a-zA-Z0-9-]+(?:\.[a-zA-Z0-9-]+)*) # FQDN
):)?(?P<port>\d+)$""", re.X)
DEFAULT_PORT = settings.BAGOU['WEBSOCKET_PORT']
DEFAULT_ADDR = settings.BAGOU['WEBSOCKET_ADDR']


class Command(BaseCommand):
    option_list = BaseCommand.option_list + (
        make_option(
            '--ipv6',
            '-6',
            action='store_true',
            dest='use_ipv6',
            default=False,
            help='Tells Tornado to use a IPv6 address.'),)
    help = "Run Tornado server with Pika Client"
    args = '[optional port number, or ipaddr:port]'

    def handle(self, addrport='', *args, **options):
        self.use_ipv6 = options.get('use_ipv6')
        if self.use_ipv6 and not socket.has_ipv6:
            raise CommandError('Your Python does not support IPv6.')
        if args:
            raise CommandError('Usage is runwebsocket %s' % self.args)
        self._raw_ipv6 = False
        if not addrport:
            self.addr = ''
            self.port = DEFAULT_PORT
        else:
            m = re.match(naiveip_re, addrport)
            if m is None:
                raise CommandError('"%s" is not a valid port number '
                                   'or address:port pair.' % addrport)
            self.addr, _ipv4, _ipv6, _fqdn, self.port = m.groups()
            if not self.port.isdigit():
                raise CommandError("%r is not a valid port number." % self.port)
            if self.addr:
                if _ipv6:
                    self.addr = self.addr[1:-1]
                    self.use_ipv6 = True
                    self._raw_ipv6 = True
                elif self.use_ipv6 and not _fqdn:
                    raise CommandError('"%s" is not a valid IPv6 address.' % self.addr)
        if not self.addr:
            self.addr = '::1' if self.use_ipv6 else DEFAULT_ADDR
            self._raw_ipv6 = bool(self.use_ipv6)

        settings.BAGOU['WEBSOCKET_PORT'] = self.port
        settings.BAGOU['WEBSOCKET_ADDR'] = self.addr
        self.run(*args, **options)

    def run(self, *args, **options):
        quit_command = 'CTRL-BREAK' if sys.platform == 'win32' else 'CONTROL-C'
        self.stdout.write((
            "%(started_at)s\n"
            "Bagou version %(version)s, using settings %(settings)r\n"
            "Starting development server at http://%(addr)s:%(port)s/\n"
            "Consuming on %(amqp_url)s (queue: %(queue_name)s)\n"
            "Quit the server with %(quit_command)s.\n"
        ) % {
            "started_at": datetime.now().strftime('%B %d, %Y - %X'),
            "version": __version__,
            "settings": settings.SETTINGS_MODULE,
            "addr": '[%s]' % self.addr if self._raw_ipv6 else self.addr,
            "amqp_url": settings.BAGOU.get('AMQP_BROKER_URL'),
            "queue_name": settings.BAGOU.get('QUEUE_NAME'),
            "port": self.port,

            "quit_command": quit_command,
        })
        websocket_server = WebSocketServer()
        websocket_server.run()



# -*- coding: utf-8 -*-
import re
import os
from setuptools import setup

README = open(os.path.join(os.path.dirname(__file__), 'README.rst')).read()

os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))


def get_version():
    VERSIONFILE = 'bagou/__init__.py'
    initfile_lines = open(VERSIONFILE, 'rt').readlines()
    VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
    for line in initfile_lines:
        mo = re.search(VSRE, line, re.M)
        if mo:
            return mo.group(1)
    raise RuntimeError('Unable to find version string in %s.' % (VERSIONFILE,))

setup(
    name='django-bagou',
    version=get_version(),
    packages=['bagou'],
    include_package_data=True,
    license='MIT License',
    description='Django Websocket for Django',
    long_description=README,
    url='https://github.com/toxinu/django-bagou',
    author='toxinu',
    author_email='toxinu@gmail.com',
    install_requires=['tornado==3.1.1', 'pika==0.9.13'],
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
)
